@@ -10,10 +10,12 @@ from dvc_data.index import (
     add,
     build,
     checkout,
+    diff,
     md5,
     read_db,
     read_json,
     save,
+    update,
     view,
     write_db,
     write_json,
@@ -162,10 +164,11 @@ def test_add(tmp_upath, as_filesystem):
     assert index[("foo",)].path == str(tmp_upath / "foo")
 
     add(index, str(tmp_upath / "data"), fs, ("data",))
-    assert len(index) == 3
+    assert len(index) == 4
     assert index[("foo",)].meta.size == 4
     assert index[("foo",)].fs == fs
     assert index[("foo",)].path == str(tmp_upath / "foo")
+    assert index[("data",)].meta.isdir
     assert index[("data", "bar")].meta.size == 4
     assert index[("data", "bar")].fs == fs
     assert index[("data", "bar")].path == str(tmp_upath / "data" / "bar")
@@ -178,12 +181,14 @@ def test_checkout(tmp_upath, odb, as_filesystem):
     index = DataIndex(
         {
             ("foo",): DataIndexEntry(
+                key=("foo",),
                 odb=odb,
                 hash_info=HashInfo(
                     name="md5", value="d3b07384d113edec49eaa6238ad5ff00"
                 ),
             ),
             ("data",): DataIndexEntry(
+                key=("data",),
                 meta=Meta(isdir=True),
                 odb=odb,
                 hash_info=HashInfo(
@@ -341,3 +346,86 @@ def test_view_traverse(odb):
         ("dir", "subdir"),
         ("dir", "subdir", "bar"),
     ]
+
+
+def test_diff():
+    from dvc_data.index.diff import ADD, DELETE, RENAME, UNCHANGED, Change
+
+    old_foo_key = ("foo",)
+    old_foo_entry = DataIndexEntry(
+        key=old_foo_key,
+        hash_info=HashInfo(
+            name="md5", value="d3b07384d113edec49eaa6238ad5ff00"
+        ),
+    )
+    old_bar_key = ("dir", "subdir", "bar")
+    old_bar_entry = DataIndexEntry(
+        key=old_bar_key,
+        hash_info=HashInfo(
+            name="md5",
+            value="1f69c66028c35037e8bf67e5bc4ceb6a.dir",
+        ),
+    )
+    old = DataIndex({old_foo_key: old_foo_entry, old_bar_key: old_bar_entry})
+
+    assert set(diff(old, old)) == {
+        Change(UNCHANGED, old_foo_entry, old_foo_entry),
+        Change(UNCHANGED, old_bar_entry, old_bar_entry),
+    }
+    assert set(diff(old, old, with_renames=True)) == {
+        Change(UNCHANGED, old_foo_entry, old_foo_entry),
+        Change(UNCHANGED, old_bar_entry, old_bar_entry),
+    }
+
+    new_foo_key = ("data", "FOO")
+    new_foo_entry = DataIndexEntry(
+        key=new_foo_key,
+        hash_info=HashInfo(
+            name="md5", value="d3b07384d113edec49eaa6238ad5ff00"
+        ),
+    )
+    new = DataIndex(
+        {
+            (
+                "data",
+                "FOO",
+            ): new_foo_entry,
+            old_bar_key: old_bar_entry,
+        }
+    )
+
+    assert set(diff(old, new)) == {
+        Change(ADD, None, new_foo_entry),
+        Change(DELETE, old_foo_entry, None),
+        Change(UNCHANGED, old_bar_entry, old_bar_entry),
+    }
+    assert set(diff(old, new, with_renames=True)) == {
+        Change(RENAME, old_foo_entry, new_foo_entry),
+        Change(UNCHANGED, old_bar_entry, old_bar_entry),
+    }
+
+
+def test_update(tmp_upath, odb, as_filesystem):
+    (tmp_upath / "foo").write_bytes(b"foo\n")
+    (tmp_upath / "data").mkdir()
+    (tmp_upath / "data" / "bar").write_bytes(b"bar\n")
+    (tmp_upath / "data" / "baz").write_bytes(b"baz\n")
+
+    fs = as_filesystem(tmp_upath.fs)
+    old = build(str(tmp_upath), fs)
+    md5(old)
+
+    index = build(str(tmp_upath), fs)
+    update(index, old)
+    assert index[("foo",)].hash_info == HashInfo(
+        "md5",
+        "d3b07384d113edec49eaa6238ad5ff00",
+    )
+    assert index[("data", "bar")].hash_info == HashInfo(
+        "md5",
+        "c157a79031e1c40f85931829bc5fc552",
+    )
+    assert index[("data", "baz")].hash_info == HashInfo(
+        "md5",
+        "258622b1688250cb619f3c9ccaefb7eb",
+    )
